@@ -19,10 +19,12 @@ DBAPI-defined Exceptions are defined in the following hierarchy::
 # Copyright (C) 2020 The Psycopg Team
 
 from typing import Any, Dict, Optional, Sequence, Tuple, Type, Union
+from typing_extensions import TypeAlias
+from asyncio import CancelledError
 
 from .pq.abc import PGconn, PGresult
 from .pq._enums import DiagnosticField
-from ._compat import TypeAlias, TypeGuard
+from ._compat import TypeGuard
 
 ErrorInfo: TypeAlias = Union[None, PGresult, Dict[int, Optional[bytes]]]
 
@@ -773,6 +775,10 @@ class SqlJsonScalarRequired(DataError,
     code='2203F', name='SQL_JSON_SCALAR_REQUIRED'):
     pass
 
+class SqlJsonItemCannotBeCastToTargetType(DataError,
+    code='2203G', name='SQL_JSON_ITEM_CANNOT_BE_CAST_TO_TARGET_TYPE'):
+    pass
+
 class FloatingPointException(DataError,
     code='22P01', name='FLOATING_POINT_EXCEPTION'):
     pass
@@ -1528,3 +1534,11 @@ class IndexCorrupted(InternalError,
 
 # autogenerated: end
 # fmt: on
+
+# Don't show a complete traceback upon raising these exception.
+# Usually the traceback starts from internal functions (for instance in the
+# server communication callbacks) but, for the end user, it's more important
+# to get the high level information about where the exception was raised, for
+# instance in a certain `Cursor.execute()`.
+
+_NO_TRACEBACK = (Error, KeyboardInterrupt, CancelledError)
